#pragma once
#include "SGString.h"
#include "PGValue.h"

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <iterator>
#include <map>
#include <string>
#include <vector>

/// @brief データ型
enum {
	null_type,
	boolean_type,
	int_type, //number_type,
	double_type,
	string_type,
	array_type,
	object_type
};

struct null {};

//-----------------------------------------------------------------------------
/**
@brief 入力
*/
class input {
protected:
	char* cur_;
	const char* end_;
	int last_ch_;
	bool ungot_;
	int line_;
public:
	input(const char* first,const char* last)
	{
		cur_ = (char*)first;
		end_ = last;
		last_ch_ = -1;
		ungot_ = false;
		line_ = 1;
	}
	int getc()
	{
		if( ungot_ ){
			ungot_ = false;
			return last_ch_;
		}
		if( cur_ == end_ ){
			last_ch_ = -1;
			return -1;
		}
		if( last_ch_ == '\n' ){
			line_++;
		}
		last_ch_ = *cur_++ & 0xff;
		return last_ch_;
	}
	int lastc()
	{
		return last_ch_;
	}
	void ungetc()
	{
		if( last_ch_ != -1 ){
			assert(!ungot_);
			ungot_ = true;
		}
	}
	char* cur() const { return cur_; }
	int line() const { return line_; }
	void skip_ws()
	{
		while(1){
			int ch = getc();
			if( !(ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r') ){
				ungetc();
				break;
			}
		}
	}
	int isget(int ch)
	{
		skip_ws();
		if( getc() != ch ){
			ungetc();
			return false;
		}
		return true;
	}
	bool match(const char* pattern)
	{
		while( *pattern ){
			if( getc() != *pattern ){
				ungetc();
				return false;
			}
			pattern++;
		}
		return true;
	}
};

//-----------------------------------------------------------------------------
/**
@brief 値
*/
class value {
public:
	typedef std::vector<value> array;
	typedef std::map<std::string, value> object;
protected:
	int type_;
	union {
		bool boolean_;
		double double_; // double number_;
		int int_;
		std::string* string_;
		array* array_;
		object* object_;
	};
public:
	value();
	value(int type, bool);
	explicit value(bool b);
	explicit value(double n);
	explicit value(int n);
	explicit value(const std::string& s);
	explicit value(const array& a);
	explicit value(const object& o);
	~value();
	value(const value& x);
	value& operator=(const value& x);
	template <typename T> bool is() const;
	template <typename T> const T& get() const;
	template <typename T> T& get();
	operator bool() const;
	const value& get(size_t idx) const;
	const value& get(const std::string& key) const;
	std::string to_str() const;
	void serialize(char* os) const;
	std::string serialize() const;

	void copy(const std::string& s, char* oi);
	void copy(const char* s, char* oi);
	void serialize_str(const std::string& s, char* oi);
	bool _parse(value& out, input& in);

	int _parse_quadhex(input &in);
	bool _parse_codepoint(std::string& out, input& in);
	bool _parse_string(value& out, input& in);
	bool _parse_string(std::string& s, input& in);
	bool _parse_array(value& out, input& in);
	bool _parse_object(value& out, input& in);
	bool _parse_number(value& out, input& in);

	char* parse(value& out, const char* first, const char* last, std::string* err);
//  std::string parse(value& out,const char* pos, const char* last);

};
typedef value::array array;
typedef value::object object;

typedef array JArray;
typedef array::const_iterator JAItetator;  ///< @brief イテレーター
typedef object::const_iterator JOCIterator; ///< @brief イテレーター

//-----------------------------------------------------------------------------
/**
@brief JSON 解析情報
*/
class JsonParser
{
public:
	JsonParser();
	virtual ~JsonParser();

	int Parse(const char* json);
	SGString Serialize();

	bool IsArray(const char* key);
	const JArray& GetArray(const char* key);
	bool GetObject(const char* key);

	bool IsString(const char* key);
	bool GetString(const char* key,SGString& str);
	bool GetInteger(const char* key,int& dat);
	bool GetBoolean(const char* key,bool& dat);
	bool GetValue(const char* key, PGValue& dat);

	JAItetator Begin(const JArray& array);
	bool Next(const JArray& array,JAItetator& it);

	JOCIterator MapBegin();
	bool MapNext( JOCIterator& it, SGString& key );

private:
	value mValue;

	const value* mValueCurrent;
};

